package edu.stanford.rsl.tutorial.book.Preprocessing;

import edu.stanford.rsl.conrad.data.numeric.Grid2D;
import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid2D;
import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid3D;
import edu.stanford.rsl.conrad.filtering.MeanFilteringTool;
import edu.stanford.rsl.conrad.utils.ImageUtil;
import ij.ImageJ;
import ij.ImagePlus;
import ij.io.Opener;

public class GuidedFilter {
	
	private  Grid2D guidance;
	private  Grid2D input;
	
	private  Grid2D result;
	
	private  float eps;
	private  float radius;

	private  MeanFilteringTool meanFilter= new MeanFilteringTool();

	public void configure(Grid2D guidance, Grid2D input){
		this.guidance = guidance;
		this.input = input;
	}
	
	public void setUp(float radius, float eps){
		this.radius = radius;
		this.eps= eps;
		meanFilter.configure((int)(radius*2), 9*(int)(radius*2));
	}
	
	public Grid2D applyToolToImage(){

			//N = boxfilter(np.ones([rows, cols]), r)
			Grid2D N = new Grid2D(input.getWidth(),input.getHeight());
			N.getGridOperator().fill(N, 1);
			N = meanFilter.applyToolToImage(N);
			
			// meanI = boxfilter(I, r) / N
			Grid2D meanI  = meanFilter.applyToolToImage(input);
			meanI.getGridOperator().divideBy(meanI, N);  
    	    //meanP = boxfilter(p, r) / N
			Grid2D meanP  = meanFilter.applyToolToImage(guidance);
			meanP.getGridOperator().divideBy(meanP, N);
			
			//meanIp = boxfilter(I * p, r) / N
			Grid2D meanIp = new Grid2D(input);
			meanIp.getGridOperator().multiplyBy(meanIp, guidance);
			meanIp = meanFilter.applyToolToImage(meanIp);
			meanIp.getGridOperator().divideBy(meanIp, N);
			
			//covIp = meanIp - meanI * meanP
			Grid2D meanIP_buf = new Grid2D(meanI);
			meanIP_buf.getGridOperator().multiplyBy(meanIP_buf,meanP);
			Grid2D covIp = new Grid2D(meanIp);
			covIp.getGridOperator().subtractBy(covIp,meanIP_buf);
			
    	    //meanII = boxfilter(I * I, r) / N
			Grid2D meanII = new Grid2D(input);
			meanII.getGridOperator().multiplyBy(meanII,input);
			meanII = meanFilter.applyToolToImage(meanII);
			meanII.getGridOperator().divideBy(meanII,N);
			
			// varI = meanII - meanI * meanI
			Grid2D meanI_buf = new Grid2D(meanI);
			meanI_buf.getGridOperator().multiplyBy(meanI_buf, meanI);
			Grid2D varI = new Grid2D(meanII);
			varI.getGridOperator().subtractBy(meanII, meanI_buf);
			
			// a = covIp / (varI + eps)
			varI.getGridOperator().addBy(varI, eps);
			Grid2D a = new Grid2D(covIp);
			a.getGridOperator().divideBy(a, varI);
			// b = meanP - a * meanI
			Grid2D b = new Grid2D(meanP);
			meanI.getGridOperator().multiplyBy(meanI, a);
			b.getGridOperator().subtractBy(b, meanI);

			//meanA = boxfilter(a, r) / N
    	    //meanB = boxfilter(b, r) / N
			a = meanFilter.applyToolToImage(a);
			b = meanFilter.applyToolToImage(b);
			a.getGridOperator().divideBy(a, N);
			b.getGridOperator().divideBy(b, N);
			
    	    //q = meanA * I + meanB
			a.getGridOperator().multiplyBy(a, input);
			a.getGridOperator().addBy(a, b);
			return a;
	}
	
	public static void main(String[] args){
		
		float eps = 0.1f*0.1f;
		float r = 0.5f;
		
		new ImageJ();
		
		String filename = "volume.tif";
		String vol ="D:\\FAU\\Projects\\Buch CT\\Messungen\\FMT\\Multitinte\\Unipapier\\Segmentierung\\60kV\\"+filename;
		
		Opener opener = new Opener();
		ImagePlus imp  = opener.openImage(vol);
		OpenCLGrid3D volume = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		OpenCLGrid3D result = new OpenCLGrid3D(volume);
		GuidedFilter gf = new GuidedFilter();
		gf.setUp(r, eps);
		long start = System.currentTimeMillis();
		for(int z = 0; z < volume.getSize()[2]; z++){

			gf.configure(volume.getSubGrid(z), volume.getSubGrid(z));
			result.setSubGrid(z, gf.applyToolToImage());
		}
		System.out.println(System.currentTimeMillis()-start);
		result.show();
	}

}
