package edu.stanford.rsl.tutorial.book.bookfitting.old;

import java.util.ArrayList;
import java.util.Arrays;

import org.mozilla.javascript.edu.emory.mathcs.backport.java.util.Collections;

import edu.stanford.rsl.conrad.data.numeric.Grid3D;
import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid3D;
import edu.stanford.rsl.conrad.utils.ImageUtil;
import ij.ImageJ;
import ij.ImagePlus;
import ij.io.Opener;
import ij.util.ArrayUtil;

public class MapBookpagesTo2D {

	//binary volume - output from algorithm
	private String dataGrid ="path_to_vol";
	//original volume
	private String multiply="path_to_vol";

	private Opener opener;
	private ImagePlus imp;
	private OpenCLGrid3D resultVolume;
	private OpenCLGrid3D data;

	//define the number of pages
	private final int numberOfPages = 10;
	
	//split pages in the center for double-written papers, false = half
	private boolean halfOrFull = true;
	long start;
	
	//load grids
	public MapBookpagesTo2D(){
		new ImageJ();
		opener = new Opener();
		imp = opener.openImage(dataGrid);
		data = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		
		imp = opener.openImage(multiply);
		OpenCLGrid3D buffer = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		
		data.getGridOperator().multiplyBy(data, buffer);
		data.setSpacing(1,1,1);
		data.setOrigin(0,0,0);
		
		extractPages();
	}

	//extracting pages
	public void extractPages(){
		//load volume
		opener = new Opener();
		imp = opener.openImage(multiply);
		OpenCLGrid3D volume = new OpenCLGrid3D(ImageUtil.wrapImagePlus(imp));
		
		if(halfOrFull){
			resultVolume = new OpenCLGrid3D( new Grid3D(volume.getSize()[2],volume.getSize()[1],numberOfPages));
		}
		else{
			resultVolume = new OpenCLGrid3D( new Grid3D(volume.getSize()[2],volume.getSize()[1],numberOfPages*2));
		}
		start = System.currentTimeMillis();
		
		//do not split the pages but treat it as only one side has writings
		if(halfOrFull)
			//walk over all pixels
			for(int z = 0; z < data.getSize()[2]; z++){
				for(int y = 0; y < data.getSize()[1]; y++){
					
					int counter = 0;
					int x = 0;
					//get the max along each page
					while(x < data.getSize()[0]){
						// if no of pages reached -> stop
						if(counter == numberOfPages)
							break;
						//else, if pixel is valid...
						if(data.getAtIndex(x, y, z) > 0){
							//write an array list, store the max and write it into resultVolume
							ArrayList<Float> values = new ArrayList<Float>();
							while(x < data.getSize()[0] && data.getAtIndex(x, y, z) > 0){
								values.add(data.getAtIndex(x, y, z));
								x++;
							}
							float max = (float)Collections.max(values);
							
							resultVolume.setAtIndex(z, y, counter++, max);
						}
						x++;
					}
				}
			}
		//Same process as above, just that we split a page in the center
		else
			for(int z = 0; z < data.getSize()[2]; z++){
				for(int y = 0; y < data.getSize()[1]; y++){
					
					int counter = 0;
					int x = 0;
					while(x < data.getSize()[0]){
						if(counter == numberOfPages*2)
							break;
						if(data.getAtIndex(x, y, z) > 0){
		
							ArrayList<Float> arrayL = new ArrayList<Float>();
							ArrayList<Float> arrayR = new ArrayList<Float>();
							while(x < data.getSize()[0] && data.getAtIndex(x, y, z) > 0){
								arrayL.add(data.getAtIndex(x, y, z));
								x++;
							}

							if(arrayL.size() > 1){
								
								for(int i = arrayL.size()/2-1; i < arrayL.size(); i++){
									arrayR.add(arrayL.get(i));
									arrayL.remove(i);
								}
	
								float max1 = (float)Collections.max(arrayL);
								float max2 = (float)Collections.max(arrayR);
								
								resultVolume.setAtIndex(z, y, counter, max1);
				
								resultVolume.setAtIndex(z, y, counter, max2);
						
							} else {
								resultVolume.setAtIndex(z, y, counter, arrayL.get(0));
							    resultVolume.setAtIndex(z, y, counter, arrayL.get(0));
							}
						}
						x++;
					}
				}
			}
		// time measure
		System.out.println(System.currentTimeMillis()-start);
		//show result
		resultVolume.show("max");
	}


	public static void main(String[] args){
		MapBookpagesTo2D  runIt = new MapBookpagesTo2D ();
	}
	
}
