package edu.stanford.rsl.tutorial.book.Preprocessing;

import edu.stanford.rsl.conrad.data.numeric.Grid2D;
import edu.stanford.rsl.conrad.data.numeric.Grid3D;
import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid2D;
import edu.stanford.rsl.conrad.data.numeric.opencl.OpenCLGrid3D;
import edu.stanford.rsl.conrad.utils.ImageUtil;
import ij.ImageJ;
import ij.ImagePlus;
import ij.io.Opener;
import ij.plugin.filter.GaussianBlur;
import ij.process.FloatProcessor;
import mpicbg.imglib.algorithm.roi.MedianFilter;
import Jama.Matrix;

public class Vesselness {

	GaussianBlur gaussian = new GaussianBlur();

	// vessel parameters
	double beta = 0.5;
	double c = 4;
	
	double b_faktor = 0;
	double c_faktor = 0;
	
	float gaussianKernelx = 1f;
	float gaussianKernely = 1f;
	
	public Vesselness(){
		b_faktor = 2 * beta * beta;
		c_faktor = 2 * c * c;
	}

	public Vesselness(float gaussianKernelx, float gaussianKernely){
		b_faktor = 2 * beta * beta;
		c_faktor = 2 * c * c;
		this.gaussianKernelx = gaussianKernelx;
		this.gaussianKernely = gaussianKernely;
	}
	
	public void setScale(float gaussianKernelx, float gaussianKernely){
		this.gaussianKernelx = gaussianKernelx;
		this.gaussianKernely = gaussianKernely;
	}
	
	public OpenCLGrid2D calculate2D(OpenCLGrid2D grid){
		
		//blur
		FloatProcessor gridProc = ImageUtil.wrapGrid2D(grid);
		gaussian.blurFloat(gridProc,gaussianKernelx,gaussianKernely,0.01);
		grid.release();
		grid = new OpenCLGrid2D(ImageUtil.wrapFloatProcessor(gridProc));
		
		OpenCLGrid2D x_f  = new OpenCLGrid2D(grid);
		OpenCLGrid2D y_f  = new OpenCLGrid2D(grid);
		OpenCLGrid2D fx  = new OpenCLGrid2D(grid);
		OpenCLGrid2D fy = new OpenCLGrid2D(grid);

		fx.getGridOperator().gradX(fx, x_f, -1, false);
		x_f.release();
		fy.getGridOperator().gradY(fy, y_f, -1, false);
		y_f.release();
		//Build Hessian
		OpenCLGrid2D fxy = new OpenCLGrid2D(fx);
		fxy.getGridOperator().gradY(fxy, fx, -1, false);
		
		OpenCLGrid2D fx2  = new OpenCLGrid2D(new Grid2D(grid.getSize()[0],grid.getSize()[1]));
		fx2.getGridOperator().gradX(fx2, fx, -1, false);
		fx.release();
		
		OpenCLGrid2D fy2  = new OpenCLGrid2D(new Grid2D(grid.getSize()[0],grid.getSize()[1]));
		fy2.getGridOperator().gradY(fy2, fy, -1, false);
		fy.release();
		
		//Normalize
		fx2.getGridOperator().multiplyBy(fx2, gaussianKernelx*gaussianKernely);
		fy2.getGridOperator().multiplyBy(fy2, gaussianKernelx*gaussianKernely);
		fxy.getGridOperator().multiplyBy(fxy, gaussianKernelx*gaussianKernely);
		
		OpenCLGrid2D ev[] = new OpenCLGrid2D[2];
		ev[0] = new OpenCLGrid2D(new Grid2D(grid.getSize()[0],grid.getSize()[1]));
		ev[1] = new OpenCLGrid2D(new Grid2D(grid.getSize()[0],grid.getSize()[1]));
		
		float maxMag = -1;

		for(int x = 0; x < grid.getSize()[0]; x++)
			for(int y = 0; y < grid.getSize()[1]; y++){
				
				double m[][] = 	{{fx2.getAtIndex(x, y),fxy.getAtIndex(x, y)},
						 		 {fxy.getAtIndex(x, y),fy2.getAtIndex(x, y)}};
				Matrix J = new Matrix(m);
				double ev_b[] =J.eig().getRealEigenvalues();
				ev[0].setAtIndex(x, y, (float)ev_b[0]);
				ev[1].setAtIndex(x, y, (float)ev_b[1]);
				float val = (float)Math.sqrt(ev_b[0]*ev_b[0]+ev_b[1]*ev_b[1]); 
				if(val > maxMag)
					maxMag = val;
			}
		
		c_faktor = 2 * maxMag/4 * maxMag/4;
		

		for(int x = 0; x < grid.getSize()[0]; x++)
			for(int y = 0; y < grid.getSize()[1]; y++)
				grid.setAtIndex(x, y, calculateVesselness2D(ev[0].getAtIndex(x, y),ev[1].getAtIndex(x, y)));
		
		/**
		 * cleanup
		 */
		fx2.release();
		fy2.release();
		fxy.release();
		ev[0].release();
		ev[1].release();
		
		return grid;
		
	}

	
	public OpenCLGrid3D calculate3D(OpenCLGrid3D grid){

		OpenCLGrid3D x_f  = new OpenCLGrid3D(grid);
		OpenCLGrid3D y_f  = new OpenCLGrid3D(grid);
		OpenCLGrid3D z_f = new OpenCLGrid3D(grid);
		OpenCLGrid3D fx  = new OpenCLGrid3D(grid);
		OpenCLGrid3D fy  = new OpenCLGrid3D(grid);
		OpenCLGrid3D fz = new OpenCLGrid3D(grid);
		
		fx.getGridOperator().gradX(fx, x_f, -1, false);
		fy.getGridOperator().gradY(fy, y_f, -1, false);
		fz.getGridOperator().gradZ(fz, z_f, -1, false);
		
		x_f.release();
		y_f.release();
		z_f.release();
		
		OpenCLGrid3D fxy = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fxy.getGridOperator().gradY(fxy, fx, -1, false);
		
		OpenCLGrid3D fxz = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fxz.getGridOperator().gradZ(fxz, fx, -1, false);

		OpenCLGrid3D fyz = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fyz.getGridOperator().gradZ(fyz, fy, -1, false);

		OpenCLGrid3D fx2 = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fx2.getGridOperator().gradX(fx2, fx, -1, false);
		fx.release();
		OpenCLGrid3D fy2 = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fy2.getGridOperator().gradY(fy2, fy, -1, false);
		fy.release();
		OpenCLGrid3D fz2 = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		fz2.getGridOperator().gradZ(fz2, fz, -1, false);
		fz.release();
		
		//Normalize
		
		fx2.getGridOperator().multiplyBy(fx2, gaussianKernelx*gaussianKernely);
		fy2.getGridOperator().multiplyBy(fy2, gaussianKernelx*gaussianKernely);
		fz2.getGridOperator().multiplyBy(fz2, gaussianKernelx*gaussianKernely);
		fxy.getGridOperator().multiplyBy(fxy, gaussianKernelx*gaussianKernely);
		fxz.getGridOperator().multiplyBy(fxz, gaussianKernelx*gaussianKernely);
		fyz.getGridOperator().multiplyBy(fyz, gaussianKernelx*gaussianKernely);
		
		OpenCLGrid3D ev[] = new OpenCLGrid3D[3];
		ev[0] = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		ev[1] = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		ev[2] = new OpenCLGrid3D(new Grid3D(grid.getSize()[0],grid.getSize()[1],grid.getSize()[2]));
		
		float maxMag = -1;

		for(int z = 0; z < grid.getSize()[2]; z++){
			for(int x = 0; x < grid.getSize()[0]; x++)
				for(int y = 0; y < grid.getSize()[1]; y++){
					
					double m[][] = 	{{fx2.getAtIndex(x, y, z),fxy.getAtIndex(x, y, z),fxz.getAtIndex(x, y, z)},
							 		{fxy.getAtIndex(x, y, z),fy2.getAtIndex(x, y, z) ,fyz.getAtIndex(x, y, z)},
							 		{fxz.getAtIndex(x, y, z),fyz.getAtIndex(x, y, z),fz2.getAtIndex(x, y, z)}};
					Matrix J = new Matrix(m);
					double ev_b[] =J.eig().getRealEigenvalues();
					ev[0].setAtIndex(x, y, z,(float)ev_b[0]);
					ev[1].setAtIndex(x, y, z,(float)ev_b[1]);
					ev[2].setAtIndex(x, y, z,(float)ev_b[2]);
					float val = (float)Math.sqrt(ev_b[0]*ev_b[0]+ev_b[1]*ev_b[1]+ev_b[2]*ev_b[2]); 
					if(val > maxMag)
						maxMag = val;
				}
		}
		
		c_faktor = 2 * maxMag/4 * maxMag/4;

		for(int z = 0; z < grid.getSize()[2]; z++)
			for(int x = 0; x < grid.getSize()[0]; x++)
				for(int y = 0; y < grid.getSize()[1]; y++){
					float val = calculateVesselness3D(ev[0].getAtIndex(x, y, z),ev[1].getAtIndex(x, y, z), ev[2].getAtIndex(x, y, z));

					if (val > 0.01)
						grid.setAtIndex(x, y,z, val);
					else
						grid.setAtIndex(x, y, z, 0);
				}
		
		return grid;
		
	}
	
	private float calculateVesselness3D(float ev1, float ev2, float ev3){

		float[] ev = eigenValuesSort(ev1,ev2,ev3);
		ev3 = ev[2];
		ev2 = ev[1];
		ev1 = ev[0];
		
		if(ev2 > 0 || ev3 > 0)
			return 0;

		else{
			if(ev2 == 0)
				ev2 = 0.00001f;
			if(ev3 == 0)
				ev3 = 0.00001f;
			double R_b = Math.abs(ev1)/Math.sqrt(Math.abs(ev2*ev3));
			double R_a = Math.abs(ev2)/Math.abs(ev3);
			double S = Math.sqrt(ev1 * ev1 + ev2* ev2 + ev3*ev3);

			return (float)( (1-Math.exp(-(R_a * R_a) / b_faktor)) * Math.exp(-(R_b * R_b) / b_faktor)* (1 - Math.exp(-(S * S)/ c_faktor)));
		}

	}
	
	private float[] eigenValuesSort(float ev1, float ev2, float ev3){
		float eigenvalues[] = new float[3];

		if(Math.abs(ev1) > Math.abs(ev2) && Math.abs(ev2) > Math.abs(ev3)){
			eigenvalues[0] = ev3;
			eigenvalues[1] = ev2;
			eigenvalues[2] = ev1;
		} else if(Math.abs(ev1) > Math.abs(ev3) && Math.abs(ev3) > Math.abs(ev2)){
			eigenvalues[0] = ev2;
			eigenvalues[1] = ev3;
			eigenvalues[2] = ev1;
		} else if(Math.abs(ev2) > Math.abs(ev3) && Math.abs(ev3) > Math.abs(ev1)){
			eigenvalues[0] = ev1;
			eigenvalues[1] = ev3;
			eigenvalues[2] = ev2;
		} else if(Math.abs(ev2) > Math.abs(ev1) && Math.abs(ev1) > Math.abs(ev3)){
			eigenvalues[0] = ev3;
			eigenvalues[1] = ev1;
			eigenvalues[2] = ev2;
		} else if(Math.abs(ev3) > Math.abs(ev1) && Math.abs(ev1) > Math.abs(ev2)){
			eigenvalues[0] = ev2;
			eigenvalues[1] = ev1;
			eigenvalues[2] = ev3;
		} else {
			eigenvalues[0] = ev1;
			eigenvalues[1] = ev2;
			eigenvalues[2] = ev3;
		}
		
		return eigenvalues;
	}
	
	private float calculateVesselness2D(float ev1, float ev2){

		if(Math.abs(ev1) > Math.abs(ev2)){
			float buf = ev2;
			ev2 = ev1;
			ev1 = buf;
		}

		if(ev2 == 0)
			ev2 = 0.00001f;
		
		double R_b = ev1 / ev2;
		double S = Math.sqrt(ev1 * ev1 + ev2* ev2);
		//check for eigenvalue > 0
		if (ev2 > 0) 
			return 0;
		else 
			return (float)(Math.exp(-(R_b * R_b) / b_faktor)* (1 - Math.exp(-(S * S)/ c_faktor)));
	}
	
	
	public static void main(String[] args){
		new ImageJ();
		
		String filename = "stackname.tif";
		String vol ="path_to_stack"+filename;
		
		Opener opener = new Opener();
		ImagePlus imp  = opener.openImage(vol);
		Vesselness vessel = new Vesselness();
		
		OpenCLGrid2D[] resultScale = new OpenCLGrid2D[imp.getStackSize()];
		for(int i = 0; i < imp.getStackSize(); i++){
			resultScale[i] = new OpenCLGrid2D(ImageUtil.wrapImagePlusSlice(imp,i+1,false));
		}

		Grid3D splitted = new Grid3D(resultScale[0].getSize()[0],resultScale[0].getSize()[1],imp.getStackSize());

		for(int z = 0; z < imp.getStackSize(); z++){
			vessel.setScale(1f, 5f);
			splitted.setSubGrid(z, vessel.calculate2D(resultScale[z]));
		}

		splitted.show("Vessel");
	}

}
